# (C) Copyright 2021 by Darren Rowland (advancedthermo.com). All rights reserved.
module AdvancedThermo

mutable struct FluidComponent
    Name::String
    Tc::Float64   # critical temperature [K]
    Pc::Float64   # critical pressure [Pa]
    ρc::Float64   # critical density [mol/m^3]
    mw::Float64   # molar mass [kg/mol]
    # polynomial-type parameters
    np::Vector{Float64}
    dp::Vector{Int64}
    tp::Vector{Float64}
    # exponential-type parameters
    ne::Vector{Float64}
    de::Vector{Int64}
    te::Vector{Float64}
    ce::Vector{Int64}
end
mutable struct FluidMixture
    components::Vector{FluidComponent}
    Tc::Vector{Float64}
    ρc::Vector{Float64}
    βv::Matrix{Float64}
    γv::Matrix{Float64}
    βT::Matrix{Float64}
    γT::Matrix{Float64}
end

function αʳ(fluid::FluidComponent,ρ,T)
    δ = ρ/fluid.ρc
    τ = fluid.Tc/T
    return αʳ(δ,τ,fluid)
end
function αʳ(m::FluidMixture,ρ,T,x)
    δ = ρ/ρᵣ(x,m.ρc,m.βv,m.γv)
    τ = Tᵣ(x,m.Tc,m.βT,m.γT)/T
    return sum(x.*αʳ.(δ,τ,m.components))
end
function αʳ(δ,τ,x,m::FluidMixture)
    return sum(x.αʳ(δ,τ,m.components))
end
function αʳ(δ,τ,f::FluidComponent)
    return αʳ(δ,τ,f.np,f.dp,f.tp) + αʳ(δ,τ,f.ne,f.de,f.te,f.ce)
end
function αʳ(δ,τ,n,d,t)
    return sum(@.(n*δ^d*τ^t))
end
function αʳ(δ,τ,n,d,t,c)
    return sum(@.(n*δ^d*τ^t*exp(-δ^c)))
end
function ρᵣ(x,ρc,βv,γv)
    return inv(vᵣ(x,inv.(ρc),βv,γv))
end
function vᵣ(x,vc,βv,γv)
    result = 0.0
    for j = 1:length(x)
        if x[j]>0.0
            for i = 1:length(x)
                if x[i]>0.0
                    vᵢⱼ = (vc[i]^(1/3)+vc[j]^(1/3))^3/8.0
                    result += ( x[i]*x[j]*βv[i,j]*γv[i,j]*
                                (x[i]+x[j])/(βv[i,j]^2*x[i]+x[j])*vᵢⱼ )
                end
            end
        end
    end
    return result
end
function Tᵣ(x,Tc,βT,γT)
    result = 0.0
    for j = 1:length(x)
        if x[j]>0.0
            for i = 1:length(x)
                if x[i]>0.0
                    Tᵢⱼ = sqrt(Tc[i]*Tc[j])
                    result += ( x[i]*x[j]*βT[i,j]*γT[i,j]*
                                (x[i]+x[j])/(βT[i,j]^2*x[i]+x[j])*Tᵢⱼ )
                end
            end
        end
    end
    return result
end
function create_mixture(fluids::Vector{FluidComponent})
    n = length(fluids)
    Tc = zeros(n)
    ρc = zeros(n)
    for i = 1:n
        Tc[i] = fluids[i].Tc
        ρc[i] = fluids[i].ρc
    end
    (βv, γv, βT, γT) = get_interaction_matrices(fluids)
    return FluidMixture(fluids, Tc, ρc, βv, γv, βT, γT)
end
function get_interaction_matrices(fluids::Vector{FluidComponent})
    n = length(fluids)
    βv = ones(n,n)
    γv = ones(n,n)
    βT = ones(n,n)
    γT = ones(n,n)
    n = length(fluids)
    # Parameters from Kunz and Wagner (2012), Table A8
    for j = 1:n
        for i = 1:n
            if fluids[i].Name=="methane" && fluids[j].Name=="helium"
                # only γv and γT non-unity
                γv[i,j] = γv[j,i] = 0.881405683
                γT[i,j] = γT[j,i] = 3.159776855
            elseif fluids[i].Name=="nitrogen" && fluids[j].Name=="helium"
                βv[i,j] = 0.969501055
                βv[j,i] = inv(βv[i,j])
                γv[i,j] = γv[j,i] = 0.932629867
                βT[i,j] = 0.692868765
                βT[j,i] = inv(βT[i,j])
                γT[i,j] = γT[j,i] = 1.471831580
            else
                # TODO(DR): more interactions as needed
            end
        end
    end
    return (βv, γv, βT, γT)
end
function get_fluid_component(name)
    if name=="methane"
        # refer Kunz & Wagner (2012) Tables A3 and A5
        np = [
            0.57335704239162
            -0.16760687523730e1
            0.23405291834916
            -0.21947376343441
            0.16369201404128e-1
            0.15004406389280e-1
        ]
        dp = [
            1
            1
            2
            2
            4
            4
        ]
        tp = [
            0.125
            1.125
            0.375
            1.125
            0.625
            1.500
        ]
        ne = [
            0.98990489492918e-1
            0.58382770929055
            -0.74786867560390
            0.30033302857974
            0.20985543806568
            -0.18590151133061e-1
            -0.15782558339049
            0.12716735220791
            -0.32019743894346e-1
            -0.68049729364536e-1
            0.24291412853736e-1
            0.51440451639444e-2
            -0.19084949733532e-1
            0.55229677241291e-2
            -0.44197392976085e-2
            0.40061416708429e-1
            -0.33752085907575e-1
            -0.25127658213357e-2
        ]
        de = [
            1
            1
            1
            2
            3
            6
            2
            3
            3
            4
            4
            2
            3
            4
            5
            6
            6
            7
        ]
        te = [
            0.625
            2.625
            2.750
            2.125
            2.000
            1.750
            4.500
            4.750
            5.000
            4.000
            4.500
            7.500
            14.000
            11.500
            26.000
            28.000
            30.000
            16.000
        ]
        ce = [
            1
            1
            1
            1
            1
            1
            2
            2
            2
            2
            2
            3
            3
            3
            6
            6
            6
            6
        ]
        return FluidComponent(
            name,
            190.564,
            0.0, # NOTE(DR) critical pressure not required (yet)
            10139.342719,
            16.04246e-3,
            np,
            dp,
            tp,
            ne,
            de,
            te,
            ce
        )
    elseif name=="nitrogen"
        # refer Kunz & Wagner (2012) Tables A3 and A5
        np = [
            0.59889711801201
            -0.16941557480731e1
            0.24579736191718
            -0.23722456755175
            0.17954918715141e-1
            0.14592875720215e-1
        ]
        dp = [
            1
            1
            2
            2
            4
            4
        ]
        tp = [
            0.125
            1.125
            0.375
            1.125
            0.625
            1.500
        ]
        ne = [
            0.10008065936206
            0.73157115385532
            -0.88372272336366
            0.31887660246708
            0.20766491728799
            -0.19379315454158e-1
            -0.16936641554983
            0.13546846041701
            -0.33066712095307e-1
            -0.60690817018557e-1
            0.12797548292871e-1
            0.58743664107299e-2
            -0.18451951971969e-1
            0.47226622042472e-2
            -0.52024079680599e-2
            0.43563505956635e-1
            -0.36251690750939e-1
            -0.28974026866543e-2
        ]
        de = [
            1
            1
            1
            2
            3
            6
            2
            3
            3
            4
            4
            2
            3
            4
            5
            6
            6
            7
        ]
        te = [
            0.625
            2.625
            2.750
            2.125
            2.000
            1.750
            4.500
            4.750
            5.000
            4.000
            4.500
            7.500
            14.000
            11.500
            26.000
            28.000
            30.000
            16.000
        ]
        ce = [
            1
            1
            1
            1
            1
            1
            2
            2
            2
            2
            2
            3
            3
            3
            6
            6
            6
            6
        ]
        return FluidComponent(
            name,
            126.192,
            0.0, # NOTE(DR) critical pressure not required (yet)
            11183.9,
            28.0134e-3,
            np,
            dp,
            tp,
            ne,
            de,
            te,
            ce
        )
    elseif name=="helium"
        # refer Kunz & Wagner (2012) Tables A4-A5
        np = [
            -0.45579024006737
            0.12516390754925e1
            -0.15438231650621e1
            0.20467489707221e-1
        ]
        dp = [
            1
            1
            1
            4
        ]
        tp = [
            0.000
            0.125
            0.750
            1.000
        ]
        ne = [
            -0.34476212380781
            -0.20858459512787e-1
            0.16227414711778e-1
            -0.57471818200892e-1
            0.19462416430715e-1
            -0.33295680123020e-1
            -0.10863577372367e-1
            -0.22173365245954e-1
        ]
        de = [
            1
            3
            5
            5
            5
            2
            1
            2
        ]
        te = [
            0.750
            2.625
            0.125
            1.250
            2.000
            1.000
            4.500
            5.000
        ]
        ce = [
            1
            1
            1
            1
            1
            2
            3
            3
        ]
        return FluidComponent(
            name,
            5.1953,
            0.0, # NOTE(DR) critical pressure not required (yet)
            17399.0,
            4.002602e-3,
            np,
            dp,
            tp,
            ne,
            de,
            te,
            ce
        )
    else
        # TODO(DR): more fluids as needed
    end
end
end # module
